#![allow(dead_code)]
pub use reqwest;
use std::fmt::{self, Display};

use serde::{Serialize, Deserialize};

#[derive(Debug)]
pub struct Context {
    client: reqwest::Client,
    api_key: Option<String>,
    url: String,

    // cache for login()
    mfa_key: Option<String>,
    device: Option<String>,
}

impl Display for Context {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        let api_key = match &self.api_key {
            Some(key) => format!("\"{}\"", key),
            None => "None".to_owned(),
        };
        writeln!(f, "Context(api_key: \"{}\", url: \"{}\")", api_key, self.url)
    }
}

#[derive(Deserialize)]
struct ErrorResponse {
    error: String,
}

#[derive(Serialize)]
struct LoginRequest<'a> {
    email: &'a str,
    password: &'a str,
    device: &'a str,
}

#[derive(Deserialize)]
struct LoginResponse {
    name: String,
    email: String,
    mfa_enabled: bool,
    mfa_key: Option<String>,
    api_key: Option<String>,
}

#[derive(Serialize)]
struct MfaRequest<'a> {
    mfa_token: &'a str,
    mfa_key: &'a str,
    device: &'a str
}

#[derive(Deserialize)]
struct MfaResponse {
    name: String,
    api_key: String,
    email: String,
}

#[derive(Serialize)]
struct UserInfoRequest<'a> {
    api_key: &'a str,
    profile_picture: Option<&'a str>,
    name: Option<&'a str>,
}

#[derive(Debug, Deserialize)]
pub struct UserInfo {
    pub name: String,
    pub is_premium: bool,
    pub email: String,
    pub in_trial: bool,
    pub profile_picture_url: Option<String>,
    pub max_alias_free_plan: u32,
}

#[derive(Debug, Deserialize)]
pub struct Stats {
    pub nb_alias: u32,
    pub nb_block: u32,
    pub nb_forward: u32,
    pub nb_reply: u32,
}

#[derive(Debug, Deserialize)]
pub struct AliasOptionsSuffix {
    pub signed_suffix: String,
    pub suffix: String,
    pub is_custom: bool,
    pub is_premium: bool,
}

#[derive(Debug, Deserialize)]
pub struct AliasOptionsRecommendation {
    pub alias: String,
    pub hostname: String,
}

#[derive(Debug, Deserialize)]
pub struct AliasOptions {
    pub can_create: bool,
    pub prefix_suggestion: String,
    pub suffixes: Vec<AliasOptionsSuffix>,
    pub recommendation: Option<AliasOptionsRecommendation>,
}

#[derive(Debug, Deserialize)]
pub struct AliasMailbox {
    pub email: String,
    pub id: u64,
}

#[derive(Debug, Deserialize)]
pub struct AliasContact {
    pub email: String,
    pub name: Option<String>,
    pub reverse_alias: String,
}

#[derive(Debug, Deserialize)]
pub struct AliasActivity {
    pub action: String,
    pub contact: AliasContact,
    pub timestamp: u64,
}

#[derive(Debug, Deserialize)]
pub struct Alias {
    pub creation_date: String,
    pub creation_timestamp: u64,
    pub email: String,
    pub name: Option<String>,
    pub enabled: bool,
    pub pinned: bool,
    pub note: Option<String>,
    pub id: u64,
    pub mailbox: AliasMailbox,
    pub mailboxes: Vec<AliasMailbox>,
    pub nb_block: u64,
    pub nb_forward: u64,
    pub nb_reply: u64,
    pub support_pgp: bool,
    pub disable_pgp: bool,
    pub latest_activity: Option<AliasActivity>,
} 

#[derive(Serialize)]
pub struct NewCustomAliasRequest<'a> {
    pub alias_prefix: &'a str,
    pub signed_suffix: &'a str,
    pub mailbox_ids: &'a [u64],
    pub note: &'a Option<&'a str>,
    pub name: &'a Option<&'a str>,
}

#[derive(Debug, Deserialize)]
pub struct AliasListResponse {
    pub aliases: Vec<Alias>,
}

pub enum NewRandomAliasMode {
    UUID, WORD,
}

pub enum AliasListMode {
    ALL,
    PINNED,
    DISABLED,
    ENABLED,
}

impl AliasListMode {
    pub fn param_name(&self) -> Option<&str> {
        match self {
            Self::ALL => None,
            Self::PINNED => Some("pinned"),
            Self::DISABLED => Some("disabled"),
            Self::ENABLED => Some("enabled"),
        }
    }
}

pub enum ErrorKind {
    State,
    Json,
    Http,
    Auth,
}

pub struct Error {
    kind: ErrorKind,
    msg: String,
}

impl Error {
    pub(crate) fn new<T>(
        kind: ErrorKind, msg: T
    ) -> Self where T: Into<String> {
        Self {kind, msg: msg.into()}
    } 

    pub fn is_state(&self) -> bool {
        matches!(self.kind, ErrorKind::State)
    }

    pub fn is_http(&self) -> bool {
        matches!(self.kind, ErrorKind::Http)
    }

    pub fn is_auth(&self) -> bool {
        matches!(self.kind, ErrorKind::Auth)
    }

    pub fn is_json(&self) -> bool {
        matches!(self.kind, ErrorKind::Json)
    }
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.msg)
    }
}

pub const DEFAULT_URL: &str = "https://app.simplelogin.io/";

impl Context {
    pub fn new(api_key: Option<&str>) -> Self {
        Self {
            client: reqwest::Client::new(),
            api_key: api_key.map(|key| key.to_owned()),
            url: DEFAULT_URL.to_owned(),
            mfa_key: None,
            device: None,
        }
    }

    pub fn new_with_url(api_key: Option<&str>, url: &str) -> Self {
        Self {
            client: reqwest::Client::new(),
            api_key: api_key.map(|key| key.to_owned()),
            url: url.to_owned(),
            mfa_key: None,
            device: None
        }
    }

    pub fn client(&self) -> &reqwest::Client {
        &self.client
    }

    pub fn api_key(&self) -> &Option<String> {
        &self.api_key
    }

    pub fn set_api_key(&mut self, api_key: Option<&str>) {
        self.api_key = api_key.map(|s| s.to_owned());
    }

    fn parse_error_json(error_body: &str) -> String {
        match serde_json::from_str(error_body) {
            Ok::<ErrorResponse, _>(obj) => obj.error,
            Err(_) => "Could not parse error response JSON".to_owned(),
        }
    }

    async fn perform_request(
        req_obj: reqwest::RequestBuilder
    ) -> Result<String, Error> {
        match req_obj.send().await {
            Ok(resp) => {
                let status = resp.status();
                let resp_text = resp.text().await
                    .map_err(|e| Error::new(ErrorKind::Http, e.to_string()))?;
                if status.is_success() {
                    Ok(resp_text)
                } else if status.is_client_error() {
                    Err(Error::new(ErrorKind::Auth, Self::parse_error_json(&resp_text)))
                } else {
                    Err(Error::new(ErrorKind::Http, status.to_string()))
                }
            },
            Err(err) => {
                Err(Error::new(ErrorKind::Http, err.to_string()))
            },
        }
    }

    async fn get_request(
        &mut self, endpoint: &str, params: &[(&str, &str)], headers: &[(&str, &str)],
    ) -> Result<String, Error> {
        let base_url = self.url.to_owned() + endpoint;
        let mut builder = self.client.get(
            reqwest::Url::parse_with_params(base_url.as_str(), params)
                .map_err(|_| {
                    Error::new(ErrorKind::Http, "Url parse error")
                })?);
        for header in headers {
            builder = builder.header(header.0, header.1);
        }
        Self::perform_request(builder).await
    }

    async fn post_request(
        &mut self, endpoint: &str, params: &[(&str, &str)],
        headers: &[(&str, &str)], body: &str
    ) -> Result<String, Error> {
        let base_url = self.url.to_owned() + endpoint;
        let mut builder = self.client.post(
            reqwest::Url::parse_with_params(base_url.as_str(), params).map_err(|_| {
                Error::new(ErrorKind::Http, "Url parse error")
            })?).header("Content-Type", "application/json")
            .body(body.to_owned());
        for header in headers {
            builder = builder.header(header.0, header.1);
        }
        Self::perform_request(builder).await
    }

    async fn patch_request(
        &mut self, endpoint: &str, body: Option<&str>,
        headers: Option<&[(&str, &str)]>
    ) -> Result<String, Error> {
        let mut builder = self.client.patch(self.url.to_owned() + endpoint);
        if let Some(body_str) = body {
            builder = builder.header("Content-Type", "application/json")
                .body(body_str.to_owned());
        }
        if let Some(headers_arr) = headers {
            for header in headers_arr {
                builder = builder.header(header.0, header.1);
            }
        }
        Self::perform_request(builder).await
    }

    pub fn obj_to_json<T>(
        obj: &T
    ) -> Result<String, Error> where T: serde::Serialize {
        serde_json::to_string(obj)
            .map_err(|e| Error::new(ErrorKind::Json, e.to_string()))
    }

    pub fn json_to_obj<'a, T>(
        json: &'a str
    ) -> Result<T, Error> where T: serde::Deserialize<'a> {
        serde_json::from_str(json)
            .map_err(|e| Error::new(ErrorKind::Json, e.to_string()))
    }

    pub async fn login(
        &mut self,
        email: &str,
        password: &str,
        device: &str
    ) -> Result<bool, Error> {
        let req_body = LoginRequest {email, password, device};
        let body = self.post_request("api/auth/login", &[], &[],
            Self::obj_to_json(&req_body)?.as_str()).await?;
        let resp_obj: LoginResponse = Self::json_to_obj(body.as_str())?;
        if resp_obj.mfa_enabled {
            if resp_obj.mfa_key.is_some() {
                self.device = Some(device.to_owned());
                self.mfa_key = resp_obj.mfa_key;
                Ok(true)
            } else {
                Err(Error::new(ErrorKind::Json, "Error response did not contain mfa key"))
            }
        } else if resp_obj.api_key.is_some() {
            self.api_key = resp_obj.api_key;
            self.mfa_key = None;
            self.device = None;
            Ok(true)
        } else {
            Err(Error::new(ErrorKind::Json, "Response did not contain api key"))
        }
    }

    pub async fn mfa(&mut self, mfa_token: &str) -> Result<(), Error> {
        if self.mfa_key.is_some() && self.device.is_some() {
            let req_json = Self::obj_to_json(&MfaRequest {
                mfa_token,
                mfa_key: self.mfa_key.as_ref().unwrap().as_str(),
                device: self.device.as_ref().unwrap().as_str(),
            })?;
            let body = self.post_request("api/auth/mfa", &[], &[], req_json.as_str()).await?; 
            let resp_obj: MfaResponse = Self::json_to_obj(body.as_str())?;
            self.api_key = Some(resp_obj.api_key);
            self.device = None;
            self.mfa_key = None;
            Ok(())
        } else {
            Err(Error::new(ErrorKind::State, "No login operation in progress"))
        }
    }

    pub async fn logout(&mut self) -> Result<(), Error> {
        match self.api_key.as_ref() {
            Some(api_key) => {
                self.get_request("api/logout", &[],
                    &[("Authentication", api_key.clone().as_str())]).await?;
                self.api_key = None;
                Ok(())
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }

    pub async fn get_user_info(&mut self) -> Result<UserInfo, Error> {
        match self.api_key.as_ref() {
            Some(api_key) => {
                let res = self.get_request("api/user_info", &[],
                    &[("Authentication", api_key.clone().as_str())]).await?;
                Ok(Self::json_to_obj(&res)?)
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }

    pub async fn update_user_info(
        &mut self, profile_picture: Option<&str>, name: Option<&str>
    ) -> Result<UserInfo, Error> {
        match self.api_key.as_ref() {
            Some(api_key) => {
                let req_json = Self::obj_to_json(
                    &UserInfoRequest {api_key, profile_picture, name})?;
                let res = self.patch_request("api/user_info",
                    Some(&req_json), None).await?;
                Ok(Self::json_to_obj(&res)?)
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }

    pub async fn stats(&mut self) -> Result<Stats, Error> {
        match self.api_key.as_ref() {
            Some(api_key) => {
                let res = self.get_request("api/stats", &[],
                    &[("Authentication", api_key.clone().as_str())]).await?;
                Ok(Self::json_to_obj(&res)?)
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }

    pub async fn alias_options(&mut self, hostname: Option<&str>) -> Result<AliasOptions, Error> {
        match self.api_key.as_ref() {
            Some(api_key) => {
                let params: Vec<(&str, &str)> = match hostname {
                    Some(hn) => vec![("hostname", hn)],
                    None => vec![],
                };
                let res = self.get_request("api/v5/alias/options", &params,
                    &[("Authentication", api_key.clone().as_str())]).await?;
                Ok(Self::json_to_obj(&res)?)
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }

    pub async fn new_custom_alias(
        &mut self, hostname: Option<&str>, alias_prefix: &str,
        signed_suffix: &str, mailbox_ids: &[u64], note: Option<&str>,
        name: Option<&str>
    ) -> Result<Alias, Error> {
        match self.api_key.as_ref() {
            Some(api_key) => {
                let params: Vec<(&str, &str)> = match hostname {
                    Some(hn) => vec![("hostname", hn)],
                    None => vec![],
                };
                let req_json = Self::obj_to_json(&NewCustomAliasRequest {
                    alias_prefix, signed_suffix, mailbox_ids,
                    name: &note, note: &name})?;
                let res = self.post_request("api/v3/alias/custom/new", &params,
                    &[("Authentication", api_key.clone().as_str())],
                    req_json.as_str()).await?;
                Ok(Self::json_to_obj(&res)?)
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }

    pub async fn new_random_alias(
        &mut self, hostname: Option<&str>,
        mode: Option<NewRandomAliasMode>, note: &str
    ) -> Result<Alias, Error> {
        match self.api_key.as_ref() {
            Some(api_key) => {
                let mut params = vec![];
                if let Some(hn) = hostname {
                    params.push(("hostname", hn));
                }
                if let Some(md) = mode {
                    params.push(("mode", match md {
                        NewRandomAliasMode::UUID => "uuid",
                        NewRandomAliasMode::WORD => "word",
                    }));
                }
                let req_json = serde_json::json!({"note": note});
                let res = self.post_request("api/alias/random/new", &params,
                    &[("Authentication", api_key.clone().as_str())],
                    req_json.as_str().unwrap()).await?;
                Ok(Self::json_to_obj(&res)?)
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }

    pub async fn list_aliases(
        &mut self, page_id: u64, mode: AliasListMode, query: Option<&str>
    ) -> Result<Vec<Alias>, Error> {
        match self.api_key.as_ref() {
            Some(api_key) => {
                let page_id_str = u64::to_string(&page_id);
                let params = match mode.param_name() {
                    Some(name) => vec![(name, "true"),
                                       ("page_id", page_id_str.as_str())],
                    None => vec![("page_id", page_id_str.as_str())],
                };
                let res = match query {
                    Some(q) => {
                        let req_json = serde_json::json!({"query": q});
                        self.post_request("api/v2/aliases", &params,
                            &[("Authentication", api_key.clone().as_str())],
                            req_json.as_str().unwrap()).await?
                    },
                    None => self.get_request("api/v2/aliases", &params,
                        &[("Authentication", api_key.clone().as_str())]).await?,
                };
                Ok(Self::json_to_obj::<AliasListResponse>(&res)?.aliases)
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }

    pub async fn get_alias(
        &mut self, alias_id: u64
    ) -> Result<Alias, Error> {
        let endpoint = format!("api/aliases/{alias_id}");
        match self.api_key.as_ref() {
            Some(api_key) => {
                let res = self.get_request(endpoint.as_str(), &[],
                    &[("Authentication", api_key.clone().as_str())]).await?;
                Ok(Self::json_to_obj(&res)?)
            },
            None => Err(Error::new(ErrorKind::State, "Not logged in")),
        }
    }
}
